import { IServiceProvider } from "./IServiceProvider.js";
import { ServiceCollection } from "./ServiceCollection.js";
import { ServiceRegistry } from "./ServiceRegistry.js";
import { getNameOfServiceType } from "../helpers/getNameOfServiceType.js";
import { ServiceInstancesStore } from "./ServiceInstancesStore.js";
import { ServiceLifetime } from "./ServiceLifetime.js";
import { isAsyncServiceImplementationLoader } from "../helpers/isAsyncServiceImplementationLoader.js";
import { validateRequestPath } from "./validation/validateRequestPath.js";
import { getNameOfDescriptor } from "../helpers/getNameOfDescriptor.js";
import { Debug } from "../debug.js";
export class ServiceProvider {
    collection;
    parent;
    static getDescriptor(provider, service) {
        return provider.getDescriptor(service);
    }
    static getDescriptors(provider, service) {
        return provider.collection.getDescriptors(service);
    }
    instancesStore;
    constructor(collection, parent) {
        this.collection = collection;
        this.parent = parent;
        this.instancesStore = new ServiceInstancesStore();
    }
    getServices(service) {
        return this.resolveGeneratorSync(this.internalGetServices(service, new Set()));
    }
    getServiceAsync(service) {
        return this.resolveGeneratorAsync(this.internalGetService(service, new Set()));
    }
    getService(service) {
        return this.resolveGeneratorSync(this.internalGetService(service, new Set()));
    }
    getServicesAsync(service) {
        return this.resolveGeneratorAsync(this.internalGetServices(service, new Set()));
    }
    createScope() {
        const serviceProvider = new ServiceProvider(this.collection, this);
        return {
            serviceProvider,
            [Symbol.dispose]: () => {
                serviceProvider[Symbol.dispose]();
            },
        };
    }
    createAsyncScope() {
        const serviceProvider = new ServiceProvider(this.collection, this);
        return {
            serviceProvider,
            [Symbol.asyncDispose]: async () => {
                await serviceProvider[Symbol.asyncDispose]();
            },
        };
    }
    *internalGetServices(service, requestedBy) {
        const services = [];
        for (const descriptor of this.collection.getDescriptors(service)) {
            services.push(yield* this.internalGetDescriptorImplementation(descriptor, requestedBy));
        }
        return services;
    }
    *internalGetService(service, requestedBy) {
        const descriptor = this.getDescriptor(service);
        return yield* this.internalGetDescriptorImplementation(descriptor, requestedBy);
    }
    *internalGetDescriptorImplementation(descriptor, requestedBy) {
        if (descriptor.lifetime === ServiceLifetime.Singleton &&
            this.parent &&
            descriptor.service !== IServiceProvider) {
            return yield* this.parent.internalGetService(descriptor.service, requestedBy);
        }
        validateRequestPath(requestedBy, descriptor);
        if (descriptor.loader) {
            return (yield [descriptor, null]);
        }
        const serviceInstance = this.instancesStore.getInstanceInfo(descriptor);
        if (serviceInstance) {
            return serviceInstance.instance;
        }
        let resolve;
        let reject;
        descriptor.loader = new Promise((res, rej) => {
            resolve = res;
            reject = rej;
        });
        descriptor.loader.catch(() => { });
        let instance;
        try {
            instance = yield* this.createInstanceFromDescriptor(descriptor, requestedBy);
        }
        catch (err) {
            reject(err);
            throw err;
        }
        finally {
            descriptor.loader = null;
        }
        try {
            return instance;
        }
        finally {
            resolve(instance);
        }
    }
    *createInstanceFromDescriptor(descriptor, requestedBy) {
        try {
            const lastRequestedBy = [...requestedBy].pop();
            if (lastRequestedBy?.lifetime !== undefined) {
                if (lastRequestedBy.lifetime === ServiceLifetime.Singleton &&
                    descriptor.lifetime === ServiceLifetime.Scoped) {
                    throw new Error(`Scoped service cannot be resolved from singleton service.`);
                }
            }
            switch (descriptor.lifetime) {
                case ServiceLifetime.Singleton: {
                    const instanceInfo = this.instancesStore.addInstance(descriptor, lastRequestedBy);
                    instanceInfo.addInstance(yield* this.initializeService(descriptor, requestedBy));
                    return instanceInfo.instance;
                }
                case ServiceLifetime.Scoped: {
                    if (!this.parent) {
                        throw new Error("Scoped services require a service scope.");
                    }
                    const instanceInfo = this.instancesStore.addInstance(descriptor, lastRequestedBy);
                    instanceInfo.addInstance(yield* this.initializeService(descriptor, requestedBy));
                    return instanceInfo.instance;
                }
                case ServiceLifetime.Transient:
                    return yield* this.initializeService(descriptor, requestedBy);
            }
        }
        catch (exception) {
            throw new Error(`Failed to initiate service ${getNameOfDescriptor(descriptor)}:\n\r${exception.message}`, { cause: exception });
        }
    }
    *initializeService(descriptor, requestedBy) {
        let implementation = descriptor.implementation;
        if (isAsyncServiceImplementationLoader(implementation)) {
            implementation = yield [descriptor, implementation];
        }
        const metadata = ServiceRegistry.get(implementation);
        if (metadata) {
            const dependencies = [];
            requestedBy = new Set([...requestedBy, descriptor]);
            for (const dependency of metadata.dependencies) {
                if (Array.isArray(dependency)) {
                    dependencies.push(yield* this.internalGetServices(dependency[0], requestedBy));
                }
                else {
                    dependencies.push(yield* this.internalGetService(dependency, requestedBy));
                }
            }
            return new implementation(...dependencies);
        }
        if (descriptor.service === IServiceProvider) {
            return this;
        }
        if (typeof implementation === "function") {
            try {
                return implementation(this);
            }
            catch (err) {
                if (err instanceof TypeError &&
                    err.message.includes("cannot be invoked without 'new'")) {
                    throw new Error(Debug.errors.classNotDecorated(implementation.name), {
                        cause: err,
                    });
                }
                else {
                    throw err;
                }
            }
        }
        return implementation;
    }
    getDescriptor(service) {
        const descriptors = this.collection.getDescriptors(service);
        if (descriptors.length === 0) {
            let name = getNameOfServiceType(service);
            const metadata = ServiceRegistry.get(service);
            if (metadata?.name) {
                name = metadata.name;
            }
            throw new Error(`No service of type "${name}" is registered`);
        }
        return descriptors[descriptors.length - 1];
    }
    resolveGeneratorSync(iterator) {
        let tempService = [];
        let error = null;
        while (true) {
            let result = null;
            if (error) {
                result = iterator.throw(error);
                error = null;
            }
            else {
                result = iterator.next(...tempService);
            }
            const { value, done } = result;
            if (done) {
                return value;
            }
            const [, loader] = value;
            if (!loader?.isLoaded()) {
                error = new Error(`Lazy service cannot be resolved synchronously`);
                continue;
            }
            tempService = [loader.getImplementation()];
        }
    }
    async resolveGeneratorAsync(iterator) {
        let tempService = [];
        let error = null;
        while (true) {
            let result = null;
            if (error) {
                result = iterator.throw(error);
                error = null;
            }
            else {
                result = iterator.next(...tempService);
            }
            try {
                const { value, done } = result;
                if (done) {
                    return value;
                }
                const [descriptor, loader] = value;
                if (loader) {
                    tempService = [await loader.load()];
                }
                else {
                    tempService = [await descriptor.loader];
                }
            }
            catch (err) {
                error = err;
            }
        }
    }
    [Symbol.dispose]() {
        this.instancesStore[Symbol.dispose]();
    }
    async [Symbol.asyncDispose]() {
        await this.instancesStore[Symbol.asyncDispose]();
    }
}
//# sourceMappingURL=ServiceProvider.js.map